"""
Mail merge module importable API.

Andrew DeOrio <awdeorio@umich.edu>
"""


from .sendmail_client import SendmailClient
from .template_message import TemplateMessage
from .exceptions import MailmergeError
